"""
The ValveYard is a class in the datenraffinerie that is responsible
for parsing the user configuration and requiring the right scan or
analysis task and passing it all necessary values from their proper
functioning

Author: Alexander Becker (a.becker@cern.ch)
Date: 2021-12-16
"""
import os
from pathlib import Path
import luigi
from .scan import Scan
from . import config_utilities as cfu
from . import control_adapter as ctrl


class ValveYard(luigi.WrapperTask):
    root_config_file = luigi.Parameter(significant=True)
    procedure_label = luigi.Parameter(significant=True)
    output_dir = luigi.Parameter(significant=True)

    def requires(self):
        """ A wrapper that parses the configuration and starts the procedure
        with the corresponding procedure label

        :raises: ConfigFormatError if either the type of the configuration entry
            does not match the allowed ones or if the YAML is malformed
        :raises: DAQConfigError if the DAQ configuration is malformed
        """
        data_dir = Path(self.output_dir)
        if not data_dir.exists():
            os.makedirs(data_dir)
        output_dir = data_dir
        procedures, workflows = cfu.parse_config_file(self.root_config_file)
        procedure_names = list(map(lambda p: p['name'], procedures))
        if self.procedure_label in procedure_names:
            procedure_index = procedure_names.index(self.procedure_label)
        else:
            raise ctrl.DAQConfigError(f"No '{self.procedure_label}' found in"
                                      " the config files")
        procedure = procedures[procedure_index]
        if procedure['type'] == 'analysis':
            raise NotImplementedError('starting analyses from the ValveYard'
                                      ' has has not been implemented yet')
        if procedure['type'] == 'daq':
            return Scan(identifier=0,
                        label=self.procedure_label,
                        output_dir=str(output_dir.resolve()),
                        output_format='hdf5',
                        scan_parameters=procedure['parameters'],
                        target_config=procedure['target_init_config'],
                        target_default_config=procedure['target_power_on_default_config'],
                        daq_system_config=procedure['daq_system_config'],
                        root_config_path=str(
                            Path(self.root_config_file).resolve()),
                        calibration=procedure['calibration'])
        raise cfu.ConfigFormatError("The type of an entry must be either "
                                    "'daq' or 'analysis'")



import sys
import click
from .valve_yard import ValveYard
from .config_utilities import ConfigFormatError
from .control_adapter import DAQConfigError
import logging
import luigi


@click.command()
@click.argument('config', type=click.Path(exists=True))
@click.argument('procedure', type=str)
@click.argument('output', type=click.Path)
@click.option('-w', '--workers', 'workers', type=int, default=4)
def cli(config, procedure, workers, output):
    try:
        run_result = luigi.build([ValveYard(
            click.format_filename(config),
            procedure,
            output)],
            local_scheduler=True,
            workers=workers,
            resources={'hexacontroller': 1}
        )
        print(run_result)
    except ConfigFormatError as err:
        print(err.message)
        sys.exit(1)
    except DAQConfigError as err:
        print("The Configuration of one of the executed"
              " DAQ procedures is malformed: ")
        print(err.message)
        sys.exit(1)
    except Exception as err:
        print("An error occured that was not properly caught")
        print(err)
        sys.exit(1)



"""
Module containing the classes that together constitute a measurement
and encapsulate the different steps needed to take a measurement using the
hexaboard
"""
from pathlib import Path
from functools import reduce
import os
import operator
import pandas as pd
import luigi
import yaml
from . import config_utilities as cfu
from . import analysis_utilities as anu
from .control_adapter import DAQSystem, TargetAdapter


class Configuration(luigi.Task):
    """
    Write the configuration file for every run this way the
    configuration can be created without needing to run the
    whole daq process
    """
    target_config = luigi.DictParameter(significant=False)
    label = luigi.Parameter(significant=True)
    output_dir = luigi.Parameter(significant=True)
    identifier = luigi.IntParameter(significant=True)
    calibration = luigi.OptionalParameter(default=None,
                                          significant=False)
    root_config_path = luigi.Parameter(significant=True)

    def requires(self):
        from valve_yard import ValveYard
        # if a calibration is needed then the delegate finding
        # the calibration and adding the subsequent tasks to the
        # to the ValveYard
        if self.calibration is not None:
            return ValveYard(self.root_config_path,
                             self.calibration,
                             str(Path(self.output_dir).resolve()))

    def output(self):
        output_path = Path(self.output_dir) / (self.label +\
            str(self.identifier) + '-config.yaml')
        return luigi.LocalTarget(output_path)

    def run(self):
        target_config = cfu.unfreeze(self.target_config)
        if self.calibration is not None:
            calib_file = self.input()[0].open('r')
            calibration = yaml.safe_load(calib_file)
            out_config = cfu.update_dict(target_config, calibration)
        else:
            out_config = target_config
        with self.output().open('w') as conf_out_file:
            yaml.dump(out_config, conf_out_file)


class Measurement(luigi.Task):
    """
    The task that performs a single measurement for the scan task
    """
    # configuration and connection to the target
    # (aka hexaboard/SingleROC tester)
    target_config = luigi.DictParameter(significant=False)

    # configuration of the (daq) system
    daq_system_config = luigi.DictParameter(significant=False)

    # Directory that the data should be stored in
    output_dir = luigi.Parameter(significant=True)
    label = luigi.Parameter(significant=True)
    identifier = luigi.IntParameter(significant=True)


    root_config_path = luigi.Parameter(significant=False)

    # calibration if one is required
    calibration = luigi.OptionalParameter(default=None,
                                          significant=False)

    resources = {'hexacontroller': 1}

    def requires(self):
        return Configuration(self.target_config,
                             self.label,
                             self.output_dir,
                             self.identifier,
                             self.calibration,
                             self.root_config_path)

    def output(self):
        """
        Specify the output of this task to be the measured data along with the
        configuration of the target during the measurement
        """
        data_path = Path(self.output_dir) / (self.label +\
            str(self.identifier) + '-data.raw')
        return (luigi.LocalTarget(data_path), self.input())

    def run(self):
        """
        Perform the measurement after configuring the different parts
        of the system
        """

        # load the possibly calibrated configuration from the configuration
        # task
        with self.input().open('r') as config_file:
            target_config = yaml.safe_load(config_file.read())
        daq_system = DAQSystem(cfu.unfreeze(self.daq_system_config))
        target = TargetAdapter(cfu.unfreeze(target_config))
        target.configure()
        data_file = self.output()[0]
        daq_system.take_data(data_file.path)


class Format(luigi.Task):
    """
    Task that unpacks the raw data into the desired data format
    also merges the yaml chip configuration with the reformatted
    data.
    """
    # configuration and connection to the target
    # (aka hexaboard/SingleROC tester)
    target_config = luigi.DictParameter(significant=False)
    target_default_config = luigi.DictParameter(significant=False)

    # configuration of the (daq) system
    daq_system_config = luigi.DictParameter(significant=False)

    # Directory that the data should be stored in
    output_dir = luigi.Parameter(significant=True)
    output_format = luigi.Parameter(significant=False)
    label = luigi.Parameter(significant=True)
    identifier = luigi.IntParameter(significant=True)

    # the path to the root config file so that the Configuration
    # task can call the valveyard if a calibration is required
    root_config_path = luigi.Parameter(True)
    # calibration if one is required
    calibration = luigi.Parameter(significant=False)

    def requires(self):
        """
        to be able to unpack the data we need the data and the
        configuration file. These are returned by the 
        """
        return (Measurement(self.target_config,
                            self.daq_system_config,
                            self.output_dir,
                            self.label,
                            self.identifier,
                            self.calibration),
                ScanConfiguration(self.output_dir,
                                  self.target_default_config))


    def output(self):
        """
        define the file that is to be produced by the unpacking step
        the identifier is used to make the file unique from the other
        unpacking steps
        """
        formatted_data_path = Path(self.output_dir) / \
            (self.label + str(self.identifier) + '-data.' + self.output_format)
        return luigi.LocalTarget(formatted_data_path.resolve())

    def run(self):
        # get the target configuration of this particular run
        # from the target power on config and the run patch
        data_file_path = Path(self.input()[0][0].path).resolve()
        with self.input()[1].open('r') as pwr_on_default_config_file:
            power_on_default_config = cfu.unfreeze(yaml.safe_load(
                pwr_on_default_config_file.read()))
        with self.input()[0][1].open('r') as run_config_file:
            run_config = cfu.unfreeze(yaml.safe_load(run_config_file.read()))
        complete_config = cfu.patch_configuration(power_on_default_config, run_config)

        # 'unpack' the data from the raw data gathered into a root file that
        # can then be merged with the configuration into a large table
        unpack_command = 'unpack'
        input_file = ' -i ' + str(data_file_path)
        data_file = self.output().path + '.root'
        output_command = ' -o ' + data_file
        output_type = ' -t unpacked'
        full_unpack_command = unpack_command + input_file + output_command\
            + output_type
        os.system(full_unpack_command)

        # load the data from the unpacked root file and merge in the data from the
        # configuration for that run with the data
        measurement_data = anu.extract_data(data_file)
        os.remove(data_file)
        measurement_data = anu.add_channel_wise_data(measurement_data, complete_config)
        measurement_data = anu.add_half_wise_data(measurement_data, complete_config)
        with self.output().temporary_path() as self.tmp_out_path:
            measurement_data.to_hdf(self.tmp_out_path, 'data')



class Scan(luigi.Task):
    """
    A Scan over one parameter or over other scans

    The scan uses the base configuration as the state of the system
    and then modifies it by applying patches constructed from
    parameter/value pairs passed to the scan and then calling either
    the measurement task or a sub-scan with the patched configurations
    as their respective base configurations
    """
    # parameters describing the position of the parameters in the task
    # tree
    identifier = luigi.IntParameter(significant=True)

    # parameters describing to the type of measurement being taken
    # and the relevant information for the measurement/scan
    label = luigi.Parameter(significant=False)
    output_dir = luigi.Parameter(significant=False)
    output_format = luigi.Parameter(significant=False)
    scan_parameters = luigi.ListParameter(significant=True)

    # configuration of the target and daq system that is used to
    # perform the scan (This may be extended with an 'environment')
    target_config = luigi.DictParameter(significant=False)
    target_default_config = luigi.DictParameter(significant=False)
    daq_system_config = luigi.DictParameter(significant=False)

    root_config_path = luigi.Parameter(significant=True)
    # calibration if one is required
    calibration = luigi.OptionalParameter(significant=False,
                                          default=None)

    supported_formats = ['hdf5']

    def requires(self):
        """
        Determine the measurements that are required for this scan to proceed.

        The Scan class is a recursive task. For every parameter(dimension) that
        is specified by the parameters argument, the scan task requires a
        set of further scans, one per value of the values entry associated with
        the parameter that the current scan is to scan over, essentially
        creating the Cartesian product of all parameters specified.
        """
        required_tasks = []
        values = self.scan_parameters[0][1]
        parameter = list(self.scan_parameters[0][0])
        target_config = cfu.unfreeze(self.target_config)
        # if there are more than one entry in the parameter list the scan still
        # has more than one dimension. So spawn more scan tasks for the lower
        # dimension
        if len(self.scan_parameters) > 1:
            # calculate the id of the task by multiplication of the length of
            # the dimensions still in the list
            task_id_offset = reduce(operator.mul,
                                    [len(param[1]) for param in
                                     self.scan_parameters[1:]])
            for i, value in enumerate(values):
                patch = cfu.generate_patch(
                            parameter, value)
                subscan_target_config = cfu.update_dict(
                        target_config,
                        patch)
                required_tasks.append(Scan(self.identifier + 1 + task_id_offset
                                           * i,
                                           self.label,
                                           self.output_dir,
                                           self.output_format,
                                           self.scan_parameters[1:],
                                           subscan_target_config,
                                           self.target_default_config,
                                           self.daq_system_config,
                                           self.root_config_path,
                                           self.calibration))
        # The scan has reached the one dimensional case. Spawn a measurement
        # for every value that takes part in the scan
        else:
            for i, value in enumerate(values):
                patch = cfu.generate_patch(parameter, value)
                measurement_config = cfu.patch_configuration(
                        target_config,
                        patch)
                required_tasks.append(Format(measurement_config,
                                             self.target_default_config,
                                             self.daq_system_config,
                                             self.output_dir,
                                             self.output_format,
                                             self.label,
                                             self.identifier + i,
                                             self.root_config_path,
                                             self.calibration))
        return required_tasks

    def run(self):
        """
        concatenate the files of a measurement together into a single file
        and write the merged data
        """
        data_seg = []
        for data_file in self.input():
            print(data_file.path)
            data_seg.append(pd.read_hdf(data_file.path))
        merged_data = pd.concat(data_seg, ignore_index=True, axis=0)
        with self.output().temporary_path() as outfile:
            merged_data.to_hdf(outfile, 'data')

    def output(self):
        """
        generate the output file for the scan task
        """
        if self.output_format in self.supported_formats:
            out_file = str(self.identifier) + 'merged.' + self.output_format
            output_path = Path(self.output_dir) / out_file
            return luigi.LocalTarget(output_path)
        raise KeyError("The output format for the scans needs to"
                       " one of the following:\n"
                       f"{self.supported_formats}")



"""
Utilities for the use with the handling of the gathered data 
in the datenraffinerie.
"""
import pandas as pd
import numpy as np
import pytest
import uproot
import os
import yaml
from pathlib import Path
import subprocess


class AnalysisError(Exception):
    def __init__(self, message):
        self.message = message


def create_measurement_dataframe(data_directory: str,
                                 channels_of_interest: list):
    """
    get the data from the *.root and the metadata from the *.yaml
    files and merge them together into a dataframe.

    This function parses and combines all the data from the root and
    yaml files into a single pandas dataframe. While performing the
    merging, only the channels selected in the *.yaml['chip_params']
    dictionary entry are selected to be merged together. All other
    channels are ignored. The selection of channels happens for each run
    so the channels of different runs may be different. All files are
    merged into a single pandas dataframe.
    """
    runs = []
    data_directory = Path(data_directory)
    rootfiles = [rf for rf in data_directory.iterdir()
            if os.path.splitext(rf)[1] == '.root']
    for rootfile in rootfiles:
        run_config_file = Path(os.path.splitext(rootfile)[0]+'.yaml')
        if run_config_file.is_file():
            run_config = extract_metadata(run_config_file)
            if run_config['keepRawData'] == 1:
                raise NotImplementedError("raw data mode not implemented")
            else:
                run_data = extract_data(rootfile)
            runs.append((run_data, run_config))
        else:
            raise AnalysisError(
                f"There was no {run_config_file} for {rootfile}")
    for i, run in enumerate(runs):
        runs[i] = merge_in_config_params(run[0], run[1])
    for i, run in enumerate(runs):
        runs[i] = split_channels(run, channels_of_interest)

    return (pd.concat([r[0] for r in runs], axis=0),  # the channels
            pd.concat([r[1] for r in runs], axis=0),  # the calib channels
            pd.concat([r[2] for r in runs], axis=0))  # the common mode chans

def split_channels(run_data, plot_channels):
    """
    split the dataset into three different sets, one for the
    normal channels, one for the calibration channels and one for the
    common mode channels.
    Also create the chip-'half' identifier in the data

    From the channels filter out the channels of interest
    """
    channel_data = pd.concat([run_data[run_data['channel'] == chan]
                             for chan in plot_channels])
    channel_data = channel_data[channel_data['channeltype'] == 0]
    channel_data.assign(half=(channel_data.channel//36)+1)
    calibration_channel_data = run_data[run_data['channeltype'] == 100]
    common_mode_data = run_data[run_data['channeltype'] == 1]
    return (channel_data, calibration_channel_data, common_mode_data)


def extract_metadata(run_config_file):
    """
    Interpret the metadata from the yaml file
    """
    with open(run_config_file, 'r') as file:
        meta_Data = yaml.safe_load(file.read())
        return meta_Data['run_params']


def extract_data(rootfile, raw_data=False):
    """
    Extract the Data from the rootfile and put it
    into a Pandas dataframe
    """
    with uproot.open(rootfile) as rfile:
        if raw_data is False:
            run_data = {}
            ttree = rfile['runsummary/summary;1']
            for cname in ttree.keys():
                run_data[cname] = pd.Series(
                    np.array(ttree[cname].array()),
                    list(range(len(ttree[cname].array()))))
            return pd.DataFrame(run_data)
        else:
            raise NotImplementedError(
                "The extraction of individual events"
                " has not been implemented")


def add_channel_wise_data(measurement_data: pd.DataFrame, complete_config: dict) -> pd.DataFrame:
    """Add channel wise data adds the data from the configuration
    to every channel that is specific to that channel

    :measurement_data: data that was gathered from the chip
    :complete_config: the complete configuration (so the default config with the
                      patch appled that is the measurement configuration
    :returns: a pandas dataframe where a column has been added for each configuration
              parameter that is channel specific the channels are determined by the
              names and types present in the measurement data.
    """
    channel_keys = roc_channel_to_dict(0, 0, 0, complete_config)
    for key in channel_keys:
        measurement_data[key] = measurement_data.apply(
                lambda x: roc_channel_to_dict(
                    x['chip'],
                    x['channel'],
                    x['channeltype'],
                    complete_config)[key], axis=1)
    return measurement_data


def add_half_wise_data(measurement_data: pd.DataFrame, complete_config: dict) -> pd.DataFrame:
    """add the config information of the chip half that corresponds to the particular
    channel

    :measurement_data: data measured from the rocs
    :complete_config: configuration of the rocs at the time of measurement.
        the half wise parameters of the rocs from this config will be added to every
        channel of the corresponding half in the `measurement_data`
    :returns: the dataframe measurement_data with added columns for the half wise
              parameters

    """
    channel_keys = roc_channel_to_globals(0, 0, 1, complete_config)
    for key in channel_keys:
        measurement_data[key] = measurement_data.apply(
                lambda x: roc_channel_to_globals(
                    x['chip'],
                    x['channel'],
                    x['channeltype'],
                    complete_config)[key], axis=1)
    return measurement_data


def add_global_data(measurement_data: pd.DataFrame, global_config: dict) -> pd.DataFrame:
    """add global Data to the dataframe. Currently there is no global data
    so there is nothing to add

    :measurement_data: the dataframe that the global parameters should be added to
    :global_config: the configuration containing the global parameter that should
        be added to the measurement dataframe (it will be added to every channel of every
        chip)
    :returns: Data with the the global parameters added to every channel currently nothing
        is done

    """
    return measurement_data

def roc_channel_to_dict(chip_id, channel_id, channel_type, complete_config):
    """Map a channel identifier to the correct part of the config

    :chip_id: chip_id from the measurement in range 0,1,2 for LD hexaboard
    :channel_id: the channel number of the channel
    :channel_type: the channel type from the measurement
    :complete_config: the complete config of the chip
    :returns: TODO

    """
    id_map = {0: 'roc_s0', 1: 'roc_s1', 2: 'roc_s2'}
    channel_type_map = {0: 'ch', 1: 'calib', 100: 'cm'}
    return complete_config[id_map[int(chip_id)]]\
        [channel_type_map[int(channel_type)]][int(channel_id)]


def roc_channel_to_globals(chip_id, channel_id, channel_type, complete_config):
    """get the chip-half wise configuration from the 

    :chip_id: TODO
    :channel_id: TODO
    :channel_type: the channel type either a normal channel.
                   calibration channel or common mode channel
    :complete_config: The complete configuration of the chip at time of measurement
    :returns: a dictionary of all global setting for the given channel

    """
    id_map = {0: 'roc_s0', 1: 'roc_s1', 2: 'roc_s2'}
    channel_type_map = {0: 'ch', 1: 'calib', 100: 'cm'}
    roc_global_keys = ['DigitalHalf', 'GlobalAnalog', 'MasterTdc', 'ReferenceVoltage']
    channel_type = channel_type_map[channel_type]
    if channel_type == 'ch':
        chip_half = 0 if channel_id < 36 else 1
    if channel_type == 'cm':
        chip_half = 0 if channel_id < 2 else 1
    if channel_type == 'calib':
        chip_half = channel_id
    roc_config = complete_config[id_map[chip_id]]
    result = {}
    for gl_key in roc_global_keys:
        for key, val in roc_config[gl_key][chip_half].items():
            result[key] = val
    return result


def merge_in_config_params(run_data: pd.DataFrame, run_config: dict):
    """
    merge the run_data with the chip parameters for that run

    Note: for the injected_channels chip parameter a column is added
          that is true when the channel was injected into and false
          otherwise.
    """
    run_params = run_config
    for param in run_params.keys():
        if param == 'injection_channels':
            injected_channels = run_params['injection_channels']
            run_data['injected'] = [y in injected_channels
                                    for y in run_data['channel']]
        elif param not in run_data.columns:
            run_param = pd.Series(run_params[param],
                                  index=run_data.index,
                                  name=param)
            run_data = pd.concat([run_data, run_param], axis=1)
    return run_data


def test_extract_data():
    test_root_path = Path('./tests/data/run.root')
    frame = extract_data(test_root_path.resolve())
    expected_columns = ['chip', 'channel', 'channeltype',
                        'adc_median', 'adc_iqr', 'tot_median',
                        'tot_iqr', 'toa_median', 'toa_iqr', 'adc_mean',
                        'adc_stdd', 'tot_mean',
                        'tot_stdd', 'toa_mean', 'toa_stdd', 'tot_efficiency',
                        'tot_efficiency_error', 'toa_efficiency',
                        'toa_efficiency_error']
    for col in expected_columns:
        _ = frame[col]


def test_add_channel_wise_data():
    """ test that all the channel wise parameters appear as a
    column in the dataframe
    """
    import config_utilities as cfu
    test_data_path = Path('./tests/data/run.root')
    frame = extract_data(test_data_path.resolve())
    configuration = cfu.load_configuration('tests/data/default.yaml')
    overlay = cfu.load_configuration('tests/data/run.yaml')
    configuration = cfu.update_dict(configuration, overlay)
    frame = add_channel_wise_data(frame, configuration)
    expected_columns = ['Adc_pedestal',
                        'Channel_off',
                        'DAC_CAL_CTDC_TOA',
                        'DAC_CAL_CTDC_TOT',
                        'DAC_CAL_FTDC_TOA',
                        'DAC_CAL_FTDC_TOT',
                        'DIS_TDC',
                        'ExtData',
                        'HZ_inv',
                        'HZ_noinv',
                        'HighRange',
                        'IN_FTDC_ENCODER_TOA',
                        'IN_FTDC_ENCODER_TOT',
                        'Inputdac',
                        'LowRange',
                        'mask_AlignBuffer',
                        'mask_adc',
                        'mask_toa',
                        'mask_tot',
                        'probe_inv',
                        'probe_noinv',
                        'probe_pa',
                        'probe_toa',
                        'probe_tot',
                        'sel_trig_toa',
                        'sel_trig_tot',
                        'trim_inv',
                        'trim_toa',
                        'trim_tot'
                       ]
    for col in expected_columns:
        print(col, frame[col])


def test_add_half_wise_data():
    """test that all the half_wise parameters are added to the dataframe
    :returns: Nothing

    """
    import config_utilities as cfu
    test_data_path = Path('./tests/data/run.root')
    frame = extract_data(test_data_path.resolve())
    default = cfu.load_configuration('tests/data/default.yaml')
    overlay = cfu.load_configuration('tests/data/run.yaml')
    config = cfu.update_dict(default, overlay)
    frame = add_half_wise_data(frame, config)
    expected_columns = [
            'Adc_TH',
            'Bx_offset',
            'CalibrationSC',
            'ClrAdcTot_trig',
            'IdleFrame',
            'L1Offset',
            'MultFactor',
            'SC_testRAM',
            'SelTC4',
            'Tot_P0',
            'Tot_P1',
            'Tot_P2',
            'Tot_P3',
            'Tot_P_Add',
            'Tot_TH0',
            'Tot_TH1',
            'Tot_TH2',
            'Tot_TH3',
            'sc_testRAM',
            'Cf',
            'Cf_comp',
            'Clr_ADC',
            'Clr_ShaperTail',
            'Delay40',
            'Delay65',
            'Delay87',
            'Delay9',
            'En_hyst_tot',
            'Ibi_inv',
            'Ibi_inv_buf',
            'Ibi_noinv',
            'Ibi_noinv_buf',
            'Ibi_sk',
            'Ibo_inv',
            'Ibo_inv_buf',
            'Ibo_noinv',
            'Ibo_noinv_buf',
            'Ibo_sk',
            'ON_pa',
            'ON_ref_adc',
            'ON_rtr',
            'ON_toa',
            'ON_tot',
            'Rc',
            'Rf',
            'S_inv',
            'S_inv_buf',
            'S_noinv',
            'S_noinv_buf',
            'S_sk',
            'SelExtADC',
            'SelRisingEdge',
            'dac_pol',
            'gain_tot',
            'neg',
            'pol_trig_toa',
            'range_indac',
            'range_inv',
            'range_tot',
            'ref_adc',
            'trim_vbi_pa',
            'trim_vbo_pa',
            'BIAS_CAL_DAC_CTDC_P_D',
            'BIAS_CAL_DAC_CTDC_P_EN',
            'BIAS_FOLLOWER_CAL_P_CTDC_EN',
            'BIAS_FOLLOWER_CAL_P_D',
            'BIAS_FOLLOWER_CAL_P_FTDC_D',
            'BIAS_FOLLOWER_CAL_P_FTDC_EN',
            'BIAS_I_CTDC_D',
            'BIAS_I_FTDC_D',
            'CALIB_CHANNEL_DLL',
            'CTDC_CALIB_FREQUENCY',
            'CTRL_IN_REF_CTDC_P_D',
            'CTRL_IN_REF_CTDC_P_EN',
            'CTRL_IN_REF_FTDC_P_D',
            'CTRL_IN_REF_FTDC_P_EN',
            'CTRL_IN_SIG_CTDC_P_D',
            'CTRL_IN_SIG_CTDC_P_EN',
            'CTRL_IN_SIG_FTDC_P_D',
            'CTRL_IN_SIG_FTDC_P_EN',
            'EN_MASTER_CTDC_DLL',
            'EN_MASTER_CTDC_VOUT_INIT',
            'EN_MASTER_FTDC_DLL',
            'EN_MASTER_FTDC_VOUT_INIT',
            'EN_REF_BG',
            'FOLLOWER_CTDC_EN',
            'FOLLOWER_FTDC_EN',
            'FTDC_CALIB_FREQUENCY',
            'GLOBAL_DISABLE_TOT_LIMIT',
            'GLOBAL_EN_BUFFER_CTDC',
            'GLOBAL_EN_BUFFER_FTDC',
            'GLOBAL_EN_TOT_PRIORITY',
            'GLOBAL_EN_TUNE_GAIN_DAC',
            'GLOBAL_FORCE_EN_CLK',
            'GLOBAL_FORCE_EN_OUTPUT_DATA',
            'GLOBAL_FORCE_EN_TOT',
            'GLOBAL_INIT_DAC_B_CTDC',
            'GLOBAL_LATENCY_TIME',
            'GLOBAL_MODE_FTDC_TOA',
            'GLOBAL_MODE_NO_TOT_SUB',
            'GLOBAL_MODE_TOA_DIRECT_OUTPUT',
            'GLOBAL_SEU_TIME_OUT',
            'GLOBAL_TA_SELECT_GAIN_TOA',
            'GLOBAL_TA_SELECT_GAIN_TOT',
            'INV_FRONT_40MHZ',
            'START_COUNTER',
            'VD_CTDC_N_D',
            'VD_CTDC_N_DAC_EN',
            'VD_CTDC_N_FORCE_MAX',
            'VD_CTDC_P_D',
            'VD_CTDC_P_DAC_EN',
            'VD_FTDC_N_D',
            'VD_FTDC_N_DAC_EN',
            'VD_FTDC_N_FORCE_MAX',
            'VD_FTDC_P_D',
            'VD_FTDC_P_DAC_EN',
            'sel_clk_rcg',
            'Calib',
            'ExtCtest',
            'IntCtest',
            'Inv_vref',
            'Noinv_vref',
            'ON_dac',
            'Refi',
            'Toa_vref',
            'Tot_vref',
            'Vbg_1v',
            'probe_dc',
            'probe_dc1',
            'probe_dc2',
           ]
    for col in expected_columns:
        print(frame[col])



"""
Module containing the adapters to the daq system consisting of the
zmq-server, zmq-client and the zmq_i2c-server. This module is and
should be the only point of interaction between the Datenraffinerie
and the daq services provided by the above mentioned programs, as
such it encapsulates the peculiarities of the underlying DAQ system
and provides a uniform API for the Datenraffinerie to use.

This file needs to be adapted if the underlying DAQ programs change
their behaviour
"""
from time import sleep
import os
import logging
import shutil
from pathlib import Path
import uuid
import zmq
import yaml
from config_utilities import diff_dict, update_dict

module_logger = logging.getLogger('hexactrl_shostnameipt.control_adapter')

class DAQError(Exception):
    def __init__(self, message):
        self.message = message


class DAQConfigError(Exception):
    def __init__(self, message):
        self.message = message


class ControlAdapter:
    """
    Class that encapsulates the configuration and communication to either
    the client or the server of the daq-system
    """

    def __init__(self, config: dict, hostname: str = None, port: str = None):
        """
        Initialize the data structure on the control computer (the one
        coordinating everything) and connect to the system component.
        Do not load any configuraion yet this is done to be able to
        load the reset / power on configuration. any change of the
        configuration after the initialisation will be written to the
        target program
        """
        self.logger = logging.getLogger(
                'hexactrl_script.contol_adapter.ControlAdapter')
        config, config_hostname, config_port = self._filter_out_network_config(
                config)
        if hostname is None:
            if config_hostname is None:
                raise DAQConfigError('No hostname given')
            hostname = config_hostname
        if port is None:
            if config_port is None:
                raise DAQConfigError('No port given')
            port = config_port
        self.hostname = hostname
        self.port = port
        self.context = zmq.Context()
        self.socket = self.context.socket(zmq.REQ)
        self.socket.connect(f"tcp://{self.hostname}:{self.port}")
        self.configuration = config
        # this is used to determin if to send the full config
        #-to the target if 'configure' is called without an argument
        self.config_written = False

    def reset(self):
        """
        reset the connection with the system component, may not reset the
        state of the component
        """
        self.socket.close()
        context = zmq.Context()
        self.socket = context.socket(zmq.REQ)
        self.socket.connect(f"tcp://{self.hostname}:{self.port}")

    def configure(self, config=None, force=False, diff=False):
        """
        send the configuration to the corresponding system component and wait
        for the configuration to be completed

        This function has to pick the right configuration to send depending on
        the force and diff flags

        Arguments:
            diff: if this flag is set, the caller knows that this is only an
            update to the configuration already on the system, so no diff
            has to be performed first. If this flag is false/unset then the
            entry is run through the cach first and only the difference of
            the config passed and the cached config is sent to the endpoint

            force: if this flag is set the cache is ignored. In this case
            the cache is updated with the new values in the config passed
            in (if there is any) and then every value in the cache is sent
            to the endpoint.
            if this flag is false/unset then the diff between the config
            passed in and the cache is sent. If no config is passed, then
            the function checks if the values of the cache have been sent
            to the endpoint yet. If the 'config_written' flag is set it exits
            otherwise it configures the endpoint with the current content
            of the cache
        """
        if config is not None:
            config, _, _= self._filter_out_network_config(config)
            if diff is False:
                if force is True:
                    self.configuration = update_dict(self.configuration, config)
                    write_config = self.configuration
                else:
                    write_config = diff_dict(self.configuration, config)
            else:
                write_config = config
        else:
            if not self.config_written or force is True:
                write_config = self.configuration
            else:
                write_config = None
        # if there is no difference between the configs simply return
        if write_config is None:
            return

        rep = self._send_and_log('configure')
        if "ready" not in rep.lower():
            raise ValueError(
                    "The configuration cannot be "
                    f" written to {self.hostname}. The target"
                    f"responded with {rep}")
        serialized_config = yaml.dump(write_config)
        self.logger.debug(f"Sending configuration:\n {serialized_config}"
                f" to {self.hostname}:{self.port}")
        self.socket.send_string(serialized_config)
        rep = self.socket.recv_string()
        self.logger.debug(f"Received string '{rep}' from {self.hostname}:{self.port}")
        if not rep == 'Configured' and not rep == 'ROC(s) CONFIGURED\n...\n':
            raise DAQError("The configuration endpoint did not indicate "
                    " a successful configuration")
        self.configuration = update_dict(self.configuration, write_config)
        self.config_written = True

    def _send_and_log(self, msg: str):
        self.logger.debug(f"Sending string '{msg}' to {self.hostname}:{self.port}")
        self.socket.send_string(msg)
        rep = self.socket.recv_string()
        self.logger.debug(f"Received string '{rep}' from {self.hostname}:{self.port}")
        return rep

    @staticmethod
    def _filter_out_network_config(config):
        """
        As there is minimal network configuration inside the daq system config
        we need to filter this out to be able to pass along the parameters that
        are intended for the actual server and client

        Also has to handle data weirdly because of technical debt in the daq c++
        software
        """
        if config is None:
            return None
        # this weird contraption needs to be build because the current zmq server
        # and client expect the ENTIRE configuration (including hexaboard and every
        # other component to be sent to them
        out_config = {}
        hostname = None
        port = None
        for key, value in config.items():
            if 'hostname' == key:
                hostname = value
            elif 'port' == key:
                port = value
            else:
                out_config[key] = value
        return out_config, hostname, port


class TargetAdapter(ControlAdapter):
    """
    The adapter that is used to control the Targets (so ROCs and
    Hexboards) currrently uses the zmq_i2c server
    """

    def __init__(self, initial_config: dict):
        """
        Initializes the target Board and loads an initial config onto it

        Arguments:
            initial_config: This config is assumed to be the goal state
            of the system, if it is passed during initialisation it is
            written to the target system
        """
        try:
            hostname = initial_config['hostname']
            port = initial_config['port']
        except KeyError as err:
            raise DAQConfigError('A hostname and port are not '
                                 ' present in any configuration'
                                 ' received') from err
        super().__init__(initial_config, hostname=hostname, port=port)
        self.logger = logging.getLogger(
                'hexactrl_script.contol_adapter.TargetAdapter')
        self.configure()

    def read_config(self, parameter: dict):
        """
        Read the values set on the ROC directly from it

        Arguments:
            paramter, dict: The parameter that should be read from the ROC

        Returns:
            the dict containing the requested parameter(s) passed in the
            parameter argument set to the value read from the ROC,
            if no parameter is passed the target configuration server will
            check which values of the configuration are cached and read
            those values from the ROC update it's cache and return the new
            values to this function which will in turn return these values
            to the caller
        """
        _ = self._send_and_log('read')
        if parameter:
            read_params = yaml.dump(parameter)
            self.logger.debug(f"Sending parameters to read:\n {read_params}"
                    f" to {self.hostname}:{self.port}")
            self.socket.send_string(read_params)
        else:
            # this reads all the values in the cache of the zmq server
            # from the roc and then returns what is in the cache
            self.logger.debug(f"Sending parameters to read:\n ''"
                    f" to {self.hostname}:{self.port}")
            self.socket.send_string("")
            read_values = self.socket.recv_string()
        self.logger.debug("Received params read from target on "
                f"{self.hostname}:{self.port}:\n"
                f"{read_values}")
        return yaml.safe_load(read_values)

    def read_pwr(self):
        # only valid for hexaboard/trophy systems
        rep = self._send_and_log('read_pwr')
        pwr = yaml.safe_load(rep)
        return pwr

    def resettdc(self):
        rep = self._send_and_log('resettdc')
        return yaml.safe_load(rep)

    def measadc(self, yamlNode: dict = None) -> dict:
        # only valid for hexaboard/trophy systems
        self.socket.send_string("measadc")
        rep = self.socket.recv_string()
        if rep.lower().find("ready") < 0:
            return
        if yamlNode is not None:
            config = yamlNode
        else:
            config = self.configuration
        self.socket.send_string(yaml.dump(config))
        rep = self.socket.recv_string()
        adc = yaml.safe_load(rep)
        return adc


class DAQAdapter(ControlAdapter):
    """
    A representation of the DAQ side of the system. It encapsulates the
    zmq-server and zmq-client
    """
    variant_key_map = {'server': 'daq', 'client': 'global'}

    def __init__(self, config: dict, variant: str):
        """
        The DAQ adapter needs to modify the configuration format of the
        Datenraffinerie to make it compatible with the current zmq-server
        and client

        Arguments:
            config, dict: The configuration of the DAQ endpoint
            variant, str: either 'server' or 'client'. lets the DAQ adapter
                make the neccesary changes to the config
        """
        self.variant = variant
        super().__init__(config)
        self.logger = logging.getLogger(
                f'hexactrl_script.contol_adapter.DAQAdapter.{self.variant}')
        self.configuration = {self.variant_key_map[self.variant]:
                              self.configuration}
        self.configure()

    def configure(self, config: dict = None):
        """
        workaround for the way the zmq-client/zmq-server handles the
        configuration together with necessary checks for the
        """
        if config is not None:
            config, _, _ = self._filter_out_network_config(config)
            super().configure({self.variant_key_map[self.variant]: config},
                              force=True)
        else:
            super().configure(force=True)

    def get_config(self):
        return self.configuration[self.variant_key_map[self.variant]]

    def start(self):
        """
        Start the aquisition of the data on the server and client
        """
        rep = ""
        while "running" not in rep.lower():
            rep = self._send_and_log('start')

    def is_done(self):
        """
        check if the current aquisition is ongoing or not
        """
        rep = self._send_and_log('run_done')
        if "notdone" in rep:
            return False
        return True

    def stop(self):
        """
        stop the currently running measurement
        """
        rep = self._send_and_log('stop')
        if not rep == 'Data puller stopped' and\
                not rep == 'Stopped':
            raise DAQError('Response of the zmq-client to'
                           " the 'stop' command invalid")

    def delay_scan(self):
        """
        perform a delay scan that tries to asses the timing conditions
        for the link between the roc and the hexacontroller
        """
        # only for daq server to run a delay scan
        rep = ""
        while "delay_scan_done" not in rep:
            rep = self._send_and_log('delayscan')


class DAQSystem:
    """
    A class that abstracts encapsulates the interactions
    with the DAQ-system (the hexacontroller, hexaboard and zmq-[server|client])

    The class implements a small two-state state machine that only allows data-taking
    via the 'take_data' function after the 'start_run' function has been called.
    The data taking is stopped via the 'stop_run' function that 
    """

    def __init__(self, daq_config):
        """
        initialise the daq system by initializing it's components (the client and
        server)
        """
        # set up the server part of the daq system (zmq-server)
        self.daq_data_base_path = None
        self.daq_data_folder = None
        if 'server' not in daq_config.keys():
            raise DAQError("There mus be a 'server' key in the initial"
                           " configuration")
        if 'client' not in daq_config.keys():
            raise DAQError("There mus be a 'client' key in the initial"
                           " configuration")
        server_config, client_config = self.split_config_into_client_and_server(
                daq_config)
        self.server = DAQAdapter(server_config, 'server')
        # set up the client part of the daq system (zmq-client)
        # the wrapping with the global needs to be done so that the client
        # accepts the configuration
        self.client = DAQAdapter(client_config, 'client')
        self._setup_data_taking_context()
        self.client.configure()

    def __del__(self):
        self.tear_down_datat_taking_context()

    @staticmethod
    def split_config_into_client_and_server(daq_config: dict):
        """
        convenience function to split the configuration passed to
        the configure method into the client and server part.

        It also compensates for all the config peculiarities of
        the daq programs in their current state.

        Should not be called by the user code!
        """
        server_config = None
        client_config = None
        if 'server' in daq_config.keys():
            server_config = daq_config['server']
        if 'client' in daq_config.keys():
            client_config = daq_config['client']
        return server_config, client_config

    def configure(self, daq_config: dict = None):
        """
        configure the daq system before starting a data-taking run.

        """
        server_config = None
        client_config = None
        if daq_config is not None:
            server_config, client_config = self.split_config_into_client_and_server(
                    daq_config)
        self.client.configure(client_config)
        self.server.configure(server_config)

    def _setup_data_taking_context(self):
        """
        setup the folders and the zmq-client configuration

        Function is called at initialisation and should not be called
        by user code

        Prepare a folder to save the raw data in and set up the client
        configuration so that the zmq-client writes into that folder
        It is expected that the folder is empty before every measurement
        as the filename of the zmq-client is not easily predictable
        """
        # get the location for the placement of the files by the
        # zmq-client, if one is already configured then use it
        # otherwise generate a new one
        client_config = self.client.get_config()
        self.run_uuid = uuid.uuid1().hex
        self.daq_data_folder = Path('/tmp') / self.run_uuid
        client_config['outputDirectory'] = str(self.daq_data_folder)
        client_config['run_type'] = self.run_uuid
        if not os.path.isdir(self.daq_data_folder):
            os.mkdir(self.daq_data_folder)

    def take_data(self, output_data_path):
        """
        function that encapsulates the data taking currently done via the
        zmq-client program. The zmq-client currently has a particular way of
        naming the files it creates that is incompatible with the way luigi
        expects the files to be named to be able to evaluate if a task has
        completed or not.

        The strategy here is to configure the zmq-client to put it's output
        into the /tmp folder of the machine running the client and the Daten-
        raffinerie and then to copy that file from the location in tmp to
        the location given by the 'output_data_path' argument of the function
        after the daq for the run has concluded
        """
        self.client.configure()
        self.client.start()
        self.server.start()
        while not self.server.is_done() or\
                len(os.listdir(self.daq_data_folder)) == 0:
            sleep(0.01)
        data_files = os.listdir(self.daq_data_folder)
        if len(data_files) > 1:
            raise DAQError("More than one file was found in the"
                           f" {self.daq_data_folder.resolve()} folder")
        data_file = data_files[0]
        shutil.move(self.daq_data_folder / data_file, output_data_path)
        self.server.stop()
        self.client.stop()

    def tear_down_datat_taking_context(self):
        """
        The complement to the 'start_run' function stops the run and cleans up
        after the run has completed
        """
        if os.path.exists(self.daq_data_folder):
            for file in self.daq_data_folder.iterdir():
                os.remove(file)
            os.rmdir(self.daq_data_folder)



from setuptools import setup

setup(
    entry_points={
        'console_scripts': [
            'datenraffinerie = datenraffinerie.datenraffinerie:cli',
        ]
    },
    install_requires=[
        'Click',
        'luigi==3.0.3',
        'pandas==1.3.5',
        'matplotlib',
        'numpy',
        'scipy',
        'uproot',
        'pyyaml',
        'zmq',
        'pytest',
    ]
)
